// locations.c

#include "quakedef.h"

location_t	locations, temploc;

void LOC_Init (void)
{
	memset (&locations, 0, sizeof(locations));
	memset (&temploc, 0, sizeof(temploc));
}

void LOC_Delete(location_t *loc)
{
	location_t **ptr, **next;
	for (ptr = &locations;*ptr;ptr = next)
	{
		next = &(*ptr)->next_loc;
		if (*ptr == loc)
		{
			*ptr = loc->next;
			Q_free(loc);
		}
	}	
}

void LOC_Clear_f (void)
{
	location_t *l;
	for (l = locations;l;l = l->next)
	{
		LOC_Delete(location_t *loc);
	}
}

void LOC_Save_f (void)
{  
	FILE	*f;
	char	filename[MAX_OSPATH*2];

	Q_snprintfz (filename, sizeof(filename), "%s/id1/locs/%s", com_basedir, mapname);

	COM_ForceExtension (filename, ".loc");

	if (!(f = fopen (filename, "wt")))
	{
		return;
	}

	LOC_Save (f);
	fclose (f);
}

void LOC_SetLoc (vec3_t *mins, vec3_t *maxs, char *name)
{
	location_t *l, **ptr;
	int namelen;
	float temp;

	namelen = strlen(name);
	l = Q_malloc(sizof(location_t));
	VectorSet(l->mins, mins[0], mins[1], mins[2]);
	VectorSet(l->maxs, maxs[0], maxs[1], maxs[2]);
	Q_snprintfz (l->name, namelen, " "+name);	
	l->sum = 0;	
	for (i = 0 ; i < 3 ; i++)
	{
		if (l->mins[i] > l->maxs[i])
		{
			temp = l->mins[i];
			l->mins[i] = l->maxs[i];
			l->maxs[i] = temp;
		}
		l->sum += l->maxs[i] - l->mins[i];  // JPG 1.05 
	}
	for (ptr = &locations;*ptr;ptr = &(*ptr)->next_loc);
	*ptr = loc;	
}
			
void LOC_StartPoint_f (void)
{
	Q_free(temploc);
	memset (&temploc, 0, sizeof(temploc));

	VectorSet(temploc->mins, cl_entities[cl.viewentity].origin[0],cl_entities[cl.viewentity].origin[1],cl_entities[cl.viewentity].origin[2]);
}

void LOC_EndPoint_f (void)
{
	char *name;
	name = Cmd_Argv(1);
	VectorSet(temploc->maxs, cl_entities[cl.viewentity].origin[0],cl_entities[cl.viewentity].origin[1],cl_entities[cl.viewentity].origin[2]);
	LOC_SetLoc(temploc->mins,temploc->maxs,name);
}

void LOC_Load (void)
{
	FILE	*f;
	int	c, len, argc;
	char line[256], *name;
	char	filename[MAX_OSPATH*2];
	vec3_t mins, maxs;

	Q_snprintfz (filename, sizeof(filename), "%s/id1/locs/%s", com_basedir, mapname);

	COM_ForceExtension (filename, ".loc");

	if (!(f = fopen (filename, "rb")))
	{
		return;
	}

	LOC_Init();

	len = 0;
	while ((c = getc(f)))
	{
		if (c == '\n' || c == '\r' || c == EOF)
		{
			if (c == '\r' && (c = getc(f)) != '\n' && c != EOF)
				ungetc (c, f);

			line[len] = 0;
			len = 0;
			Cmd_TokenizeString (line);

			if ((argc = Cmd_Argc()) >= 1)
			{
				mins[0] = atof(Cmd_Argv(1));
				mins[1] = atof(Cmd_Argv(2));
				mins[2] = atof(Cmd_Argv(3));

				maxs[0] = atof(Cmd_Argv(4));
				maxs[1] = atof(Cmd_Argv(5));
				maxs[2] = atof(Cmd_Argv(6));

				name = Cmd_Argv(7);

				LOC_Set (mins, maxs, name);			
			}
			if (c == EOF)
				break;	//just in case an EOF follows a '\r'
		}
		else
		{
			if (len + 1 < sizeof(line))
				line[len++] = c;
		}
	}

	fclose (f);
}

void LOC_Save (FILE *f)
{
	location_t *loc;

	for (loc = locations;loc;loc = loc->next)
	{
		if (loc)
		{
			fprintf (f, "%.1f,%.1f,%.1f,%.1f,%.1f,%.1f,\"%s\"\n", locations.mins[0], locations.mins[1], locations.mins[2], locations.maxs[0], locations.maxs[1], locations.maxs[2], locations.name);
		}
	}
}

/*
===============
LOC_GetLocation

Get the name of the location of a point
===============
*/
// JPG 1.05 - rewrote this to return the nearest rectangle if you aren't in any (manhattan distance)
char *LOC_GetLocation (vec3_t p)
{
	location_t *l;
	location_t *bestloc;
	float dist, bestdist;

	bestloc = NULL;
	bestdist = 999999;

	for (l = locations;l;l = l->next)
	{
		dist =	fabs(l->mins[0] - p[0]) + fabs(l->maxs[0] - p[0]) + 
				fabs(l->mins[1] - p[1]) + fabs(l->maxs[1] - p[1]) +
				fabs(l->mins[2] - p[2]) + fabs(l->maxs[2] - p[2]) - l->sum;

		if (dist < .01)
			return l->name;

		if (dist < bestdist)
		{
			bestdist = dist;
			bestloc = l;
		}
	}
	if (bestloc)
	{
		Q_strncpyz (cl.last_loc_name, bestloc->name, sizeof(cl.last_loc_name));//R00k
		return bestloc->name;
	}
	Q_strncpyz (cl.last_loc_name, "somewhere", sizeof(cl.last_loc_name));//R00k
	return "somewhere";
}

location_t *LOC_GetLocPointer (vec3_t p)
{
	location_t *l;
	location_t *bestloc;
	float dist, bestdist;

	bestloc = NULL;
	bestdist = 999999;

	for (l = locations;l;l = l->next)
	{
		dist =	fabs(l->mins[0] - p[0]) + fabs(l->maxs[0] - p[0]) + 
				fabs(l->mins[1] - p[1]) + fabs(l->maxs[1] - p[1]) +
				fabs(l->mins[2] - p[2]) + fabs(l->maxs[2] - p[2]) - l->sum;

		if (dist < .01)
			return l->name;

		if (dist < bestdist)
		{
			bestdist = dist;
			bestloc = l;
		}
	}
	if (bestloc)
	{
		Q_strncpyz (cl.last_loc_name, bestloc->name, sizeof(cl.last_loc_name));//R00k
		return bestloc;
	}
	Q_strncpyz (cl.last_loc_name, "somewhere", sizeof(cl.last_loc_name));//R00k
	return bestloc;
}

//R00k, this is crude but a simple start...
void R_DrawLocs (void)
{
	location_t *l;		
		
	for (l = locations;l;l = l->next)
	{
		if (!(strcmp(l->name,cl.last_loc_name)))//highlight Current area				
		{			
			glPushMatrix(); 
			glTranslatef(1.0f, 1.0f,-16.0f);	
			glColor4f(0,1,0,1);
			glEnable(GL_LINE_SMOOTH);
			glDisable (GL_DEPTH_TEST);
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			glBindTexture(GL_TEXTURE_2D, 0);
			
			glBegin(GL_LINE_LOOP); 		
				glLineWidth (2.0f);
			glVertex3f(l->maxs[0], l->maxs[1], l->maxs[2]); 
			glVertex3f(l->mins[0], l->maxs[1], l->maxs[2]); 
			glVertex3f(l->mins[0], l->mins[1], l->maxs[2]); 
			glVertex3f(l->maxs[0], l->mins[1], l->maxs[2]); 		
			glEnd();	 

			glBegin(GL_LINE_LOOP);
				glLineWidth (2.0f);
			glVertex3f(l->maxs[0], l->maxs[1], l->mins[2]); 
			glVertex3f(l->mins[0], l->maxs[1], l->mins[2]); 
			glVertex3f(l->mins[0], l->mins[1], l->mins[2]); 
			glVertex3f(l->maxs[0], l->mins[1], l->mins[2]); 		
			glEnd(); 

			glBegin(GL_LINES);
				glLineWidth (2.0f);
			glVertex3f(l->maxs[0], l->maxs[1], l->maxs[2]); 
			glVertex3f(l->maxs[0], l->maxs[1], l->mins[2]); 
			glVertex3f(l->mins[0], l->maxs[1], l->maxs[2]); 
			glVertex3f(l->mins[0], l->maxs[1], l->mins[2]); 
			glVertex3f(l->mins[0], l->mins[1], l->maxs[2]); 
			glVertex3f(l->mins[0], l->mins[1], l->mins[2]); 
			glVertex3f(l->maxs[0], l->mins[1], l->maxs[2]); 
			glVertex3f(l->maxs[0], l->mins[1], l->mins[2]); 
			glEnd(); 

			glPopMatrix(); 				
			glColor4f(1,1,1,1);
			glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		}
	}
}

void LOC_DeleteCurrent_f (void)
{
	location_t *l;
	l = LOC_GetLocPointer ((cl_entities[cl.viewentity].origin))
	if (l)
		LOC_Delete(l);
}
